/*
 * citmail.c v3.1
 *
 * This program may be used as a local mail delivery agent, which will allow
 * all Citadel users to receive Internet e-mail.  To enable this functionality,
 * you must tell sendmail, smail, or whatever mailer you are using, that this
 * program is your local mail delivery agent.  This program is a direct
 * replacement for lmail, deliver, or whatever.
 *
 */

/* Where to find the alias table */
#define ALIASES			"./network/mail.aliases"

/* Where to place uuencoded files mailed to 'uudecode@yourhost' */
#define LOADINGDOCK		"./files/bitbucket"

/* Where to place uuencoded cit86net spool mailed to 'cit86net@yourhost' */
#define CIT86NET		"./network/spoolin.cit86net"

/* How to transmit e-mail (in case we need to bounce a message) */
#define SENDMAIL		"sendmail %s"

/* Fallback mailer for delivery to /etc/passwd users */
#define FALLBACK		"lmail %s"	

#include "sysdep.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <pwd.h>
#include <errno.h>
#include "citadel.h"

#ifdef USE_SYSLOG
#include <syslog.h>
#endif 

#define LOCAL 0
#define REMOTE 1
#define UUCP 2
#define CCITADEL 3

#undef tolower
#define tolower(x) isupper(x) ? (x+'a'-'A') : x

void get_config();
struct config config;
char nodename[16];

char *monthdesc[] = {	"Jan","Feb","Mar","Apr","May","Jun",
			"Jul","Aug","Sep","Oct","Nov","Dec" };

long conv_date(sdbuf)
char sdbuf[]; {
	int a,b,cpos,tend,tval;
	long now;
	struct tm *tmbuf;
	char dbuf[128];

	strcpy(dbuf,sdbuf);
	time(&now);
	tmbuf = (struct tm *)localtime(&now);

	/* get rid of + or - timezone mods */
	for (a=0; a<strlen(dbuf); ++a) 
		if ((dbuf[a]=='+')||(dbuf[a]=='-'))
			do {
				strcpy(&dbuf[a],&dbuf[a+1]);
				} while ((dbuf[a]!=32)&&(dbuf[a]!=0));

	/* try and extract the time by looking for colons */
	cpos = (-1);
	for (a=strlen(dbuf); a>=0; --a)
		if ((dbuf[a]==':')&&(atoi(&dbuf[a-1])!=0)) cpos=a;
	if (cpos>=0) {
		cpos = cpos - 2;
		tend = strlen(dbuf);
		for (a=tend; a>=cpos; --a) if (dbuf[a]==' ') tend=a;

		tmbuf->tm_hour = atoi(&dbuf[cpos]);
		tmbuf->tm_min = atoi(&dbuf[cpos+3]);
		tmbuf->tm_sec = atoi(&dbuf[cpos+6]);

		do {
			strcpy(&dbuf[cpos],&dbuf[cpos+1]);
			} while ((dbuf[cpos]!=32)&&(dbuf[cpos]!=0));
		}

	/* next try to extract a month */
	
	tval = (-1);
	for (a=0; a<strlen(dbuf); ++a)
		for (b=0; b<12; ++b)
			if (!strncmp(&dbuf[a],monthdesc[b],3)) {
				tval = b;
				cpos = a;
				}
	if (tval >= 0) {
		tmbuf->tm_mon = tval;
		strcpy(&dbuf[cpos],&dbuf[cpos+3]);
		}

	/* now the year */

	for (a=0; a<strlen(dbuf); ++a)
		if ((atoi(&dbuf[a])>=1900) && (dbuf[a]!=32)) {
			tmbuf->tm_year = atoi(&dbuf[a]) - 1900;
			strcpy(&dbuf[a],&dbuf[a+4]);
			}

	/* whatever's left is the mday (hopefully) */

	for (a=0; a<strlen(dbuf); ++a)
		if ((dbuf[a]!=32)&&(atoi(&dbuf[a])>=1)&&(atoi(&dbuf[a])<=31)
		   && ( (a==0)||(dbuf[a-1]==' ') ) ) {
			tmbuf->tm_mday = atoi(&dbuf[a]);
			strcpy(&dbuf[a],&dbuf[a+2]);
			}

	return((long)mktime(tmbuf));
	}


int struncmp(lstr,rstr,l)
char lstr[],rstr[]; {
	int pos = 0;
	char lc,rc;
	while (1) {
		if (pos==l) return(0);
		lc=tolower(lstr[pos]);
		rc=tolower(rstr[pos]);
		if ((lc==0)&&(rc==0)) return(0);
		if (lc<rc) return(-1);
		if (lc>rc) return(1);
		pos=pos+1;
		}
	}


#ifdef NO_STRERROR
/*
 * replacement strerror() for systems that don't have it
 */
char *strerror(e)
int e; {
	static char buf[32];

	sprintf(buf,"errno = %d",e);
	return(buf);
	}
#endif

int haschar(st,ch)
char st[];
int ch; {
	int a,b;
	b=0;
	for (a=0; a<strlen(st); ++a) if (st[a]==ch) ++b;
	return(b);
	}

void strip_trailing_whitespace(buf)
char buf[]; {
	while(isspace(buf[strlen(buf)-1]))
		buf[strlen(buf)-1]=0;
	}

int islocalok(recp,posbuf)
char recp[];
long *posbuf; {
	struct usersupp ust;
	long lookfor;
	char a_recp[128];
	int found_closest_match = 0;
	int a,us;
	strcpy(a_recp,recp);
	for (a=0; a<strlen(a_recp); ++a)
		if (a_recp[a]=='_') a_recp[a]=32;
	lookfor = (-1L); if (!struncmp(recp,"cit",3)) lookfor=atol(&recp[3]);
	us=open("/appl/citadel/usersupp",O_RDONLY);
	*posbuf = 0L;
	if (us>=0) {
		while(read(us,&ust,sizeof(struct usersupp))>0) {
			if (lookfor == ust.eternal) {
				strcpy(recp,ust.fullname);
				close(us);
				return(2);
				}
			if (!strucmp(ust.fullname,a_recp)) {
				strcpy(recp,ust.fullname);
				close(us);
				return(3);
				}
			if (!struncmp(ust.fullname,a_recp,strlen(a_recp))) {
				strcpy(recp,ust.fullname);
				found_closest_match = 1;
				}
			*posbuf = (*posbuf)+(long)sizeof(struct usersupp);
			}
		close(us);
		}
	if (getpwnam(recp)!=NULL) return(1);
	if (found_closest_match) return(3);
	return(0);
	}

/* strip leading and trailing spaces */
void striplt(buf)
char buf[]; {
	while ( (strlen(buf)>0) && (buf[0]==32) ) strcpy(buf,&buf[1]);
	while (buf[strlen(buf)-1] == 32) buf[strlen(buf)-1] = 0;
	}

void procfrom(from,user,node,name)
char from[];
char user[];
char node[];
char name[];  {
	int a;

	/* extract full name - first, it's From minus <userid> */
	strcpy(name,from);
	for (a=0; a<strlen(name); ++a) if (name[a]=='<') {
		do {
			strcpy(&name[a],&name[a+1]);
			} while ( (strlen(name) > 0) && (name[a]!='>') );
		strcpy(&name[a],&name[a+1]);
		}

	/* strip anything to the left of a bang */
	while ( (strlen(name)>0) && (haschar(name,'!')>0) ) 
		strcpy(name,&name[1]);

	/* and anything to the right of a @ or % */
	for (a=0; a<strlen(name); ++a) {
		if (name[a]=='@') name[a]=0;
		if (name[a]=='%') name[a]=0;
		}

	/* but if there are parentheses, that changes the rules... */
	if ( (haschar(from,'(') == 1) && (haschar(from,')') == 1) ) {
		strcpy(name,from);
		while ( (strlen(name) > 0) && (name[0]!='(') ) {
			strcpy(&name[0],&name[1]);
			}
		strcpy(&name[0],&name[1]);
		for (a=0; a<strlen(name); ++a)
			 if (name[a]==')') name[a]=0;
		}

	/* but if there are a set of quotes, that supersedes everything */
	if (haschar(from,34)==2) {
		strcpy(name,from);
		while ( (strlen(name) > 0) && (name[0]!=34) ) {
			strcpy(&name[0],&name[1]);
			}
		strcpy(&name[0],&name[1]);
		for (a=0; a<strlen(name); ++a)
			if (name[a]==34) name[a]=0;
		}

	/* extract user id */
	strcpy(user,from);

	/* first get rid of anything in parens */
	for (a=0; a<strlen(user); ++a) if (user[a]=='(') {
		do {
			strcpy(&user[a],&user[a+1]);
			} while ( (strlen(user) > 0) && (user[a]!=')') );
		strcpy(&user[a],&user[a+1]);
		}

	/* if there's a set of angle brackets, strip it down to that */
	if ( (haschar(user,'<') == 1) && (haschar(user,'>') == 1) ) {
		while ( (strlen(user) > 0) && (user[0]!='<') ) {
			strcpy(&user[0],&user[1]);
			}
		strcpy(&user[0],&user[1]);
		for (a=0; a<strlen(user); ++a)
			 if (user[a]=='>') user[a]=0;
		}

	/* strip anything to the left of a bang */
	while ( (strlen(user)>0) && (haschar(user,'!')>0) ) 
		strcpy(user,&user[1]);

	/* and anything to the right of a @ or % */
	for (a=0; a<strlen(user); ++a) {
		if (user[a]=='@') user[a]=0;
		if (user[a]=='%') user[a]=0;
		}


	
	/* extract node name */
	strcpy(node,from);

	/* first get rid of anything in parens */
	for (a=0; a<strlen(node); ++a) if (node[a]=='(') {
		do {
			strcpy(&node[a],&node[a+1]);
			} while ( (strlen(node) > 0) && (node[a]!=')') );
		strcpy(&node[a],&node[a+1]);
		}

	/* if there's a set of angle brackets, strip it down to that */
	if ( (haschar(node,'<') == 1) && (haschar(node,'>') == 1) ) {
		while ( (strlen(node) > 0) && (node[0]!='<') ) {
			strcpy(&node[0],&node[1]);
			}
		strcpy(&node[0],&node[1]);
		for (a=0; a<strlen(node); ++a)
			 if (node[a]=='>') node[a]=0;
		}

	/* strip anything to the left of a @ */
	while ( (strlen(node)>0) && (haschar(node,'@')>0) ) 
		strcpy(node,&node[1]);

	/* strip anything to the left of a % */
	while ( (strlen(node)>0) && (haschar(node,'%')>0) ) 
		strcpy(node,&node[1]);

	/* reduce multiple system bang paths to node!user */
	while ( (strlen(node)>0) && (haschar(node,'!')>1) ) 
		strcpy(node,&node[1]);

	/* now get rid of the user portion of a node!user string */
	for (a=0; a<strlen(node); ++a) if (node[a]=='!') node[a]=0;



	/* strip leading and trailing spaces in all strings */
	striplt(user);
	striplt(node);
	striplt(name);
	}

void bounce(from,to)	/* bounce a message back to sender */
char *from,*to; {
	static char rmcmd[512];
	FILE *rmail;

	sprintf(rmcmd,SENDMAIL,from);
	strcat(rmcmd,"\n");
	rmail=popen(rmcmd,"w");
	if (rmail==NULL) return;
	fprintf(rmail,"From postmaster\n");
	fprintf(rmail,"From: Local Mail Subsystem <postmaster@%s>\n",FQDN);
	fprintf(rmail,"To: %s\n",from);
	fprintf(rmail,"Subject: failed mail\n");
	fprintf(rmail,"\n\n%s could not deliver your mail to:\n",FQDN);
	fprintf(rmail,"%s\n",to);
	fprintf(rmail,"because there is no such user on this system.\n");
	fprintf(rmail,"\n(Unsent message does *not* follow.  ");
	fprintf(rmail,"Help to conserve bandwidth.)\n");
	pclose(rmail);
	}


void do_citmail(from,recp,pos,dtype)  /* pipe message through netproc */
char from[];
char recp[];
long pos;
int dtype; {
	long now;
	FILE *temp;
	int a;
	char buf[128];
	char userbuf[128];
	char frombuf[128];
	char nodebuf[128];
	char destsys[128];
	char subject[128];

	if (dtype==REMOTE) {

		/* get the Citadel node name out of the path */
		strncpy(destsys,recp,sizeof(destsys));
		for (a=0; a<strlen(destsys); ++a)
			if ((destsys[a]=='!')||(destsys[a]=='.'))
				destsys[a]=0;

		/* chop the system name out, so we're left with a user */
		while (haschar(recp,'!')) strcpy(recp,&recp[1]);

		/* now convert underscores to spaces */
		for (a=0; a<strlen(recp); ++a) if (recp[a]=='_') recp[a]=' ';

		}

	time(&now);
	for (a=0; a<strlen(from); ++a) if (from[a]==32) from[a]=0;

	sprintf(buf,"./network/spoolin/citmail.%d",getpid());
	temp = fopen(buf,"w");

	putc(255,temp); putc(MES_NORMAL,temp); putc(1,temp);
	procfrom(from,userbuf,nodebuf,frombuf);
	strcpy(subject,"");
	strcpy(nodebuf,nodename);
	while ( (fgets(buf,128,stdin)!=NULL) && (buf[1]!=0) ) {
		strip_trailing_whitespace(buf);
		if (!strncmp(buf,"Subject: ",9)) strcpy(subject,&buf[9]);
		if (!strncmp(buf,"Date: ",6)) now = conv_date(&buf[6]);
		if (!strncmp(buf,"From: ",6)) {
			procfrom(&buf[6],userbuf,nodebuf,frombuf);
			}
		}

	/* now convert it to Citadel format */
	fprintf(temp,"P%s@%s",userbuf,nodebuf);	putc(0,temp);
	fprintf(temp,"T%ld",now);		putc(0,temp);
	fprintf(temp,"A%s",frombuf);		putc(0,temp);
	fprintf(temp,"OMail");			putc(0,temp);
	fprintf(temp,"N%s",nodebuf);		putc(0,temp);
	if (dtype==REMOTE) {
		fprintf(temp,"D%s",destsys); putc(0,temp);
		}
	fprintf(temp,"R%s",recp); putc(0,temp);
	if (strlen(subject)>0) {
		fprintf(temp,"U%s",subject); putc(0,temp);
		}
	putc('M',temp);
	/* fprintf(temp,"Path: %s\n",from); */
	while (a=getc(stdin), a>0) putc(a,temp);
	putc(0,temp);
	fclose(temp);
	}

void do_roommail(from,recp)  /* pipe public message through netproc */
char from[];
char recp[]; {
	long now;
	FILE *temp;
	int a;
	char buf[128],userbuf[128],frombuf[128],nodebuf[128];
	char subject[128];

	strcpy(subject,"");
	strcpy(recp,&recp[5]);
	for (a=0; a<strlen(recp); ++a) if (recp[a]=='_') recp[a]=32;
	time(&now);
	for (a=0; a<strlen(from); ++a) if (from[a]==32) from[a]=0;


	sprintf(buf,"./network/spoolin/citmail.%d",getpid());
	temp = fopen(buf,"w");

	putc(255,temp); putc(MES_NORMAL,temp); putc(1,temp);
	strcpy(frombuf,"UUCP Mailer");
	strcpy(nodebuf,nodename);
	while ( (fgets(buf,128,stdin)!=NULL) && (buf[1]!=0) ) {
		strip_trailing_whitespace(buf);
		if (!strncmp(buf,"Subject: ",9)) strcpy(subject,&buf[9]);
		if (!strncmp(buf,"Date: ",6)) now = conv_date(&buf[6]);
		if (!strncmp(buf,"From: ",6)) {
			procfrom(&buf[6],userbuf,nodebuf,frombuf);
			}
		}

	fprintf(temp,"P%s@%s",userbuf,nodebuf);	putc(0,temp);
	fprintf(temp,"T%ld",now); putc(0,temp);
	fprintf(temp,"A%s",frombuf); putc(0,temp);
	fprintf(temp,"O%s",recp); putc(0,temp);
	fprintf(temp,"N%s",nodebuf); putc(0,temp);
	if (strlen(subject)>0) {
		fprintf(temp,"U%s",subject); putc(0,temp);
		}
	putc('M',temp);
	while (a=getc(stdin), a>0) putc(a,temp);
	putc(0,temp);
	fclose(temp);
	}

void do_uudecode(target)
char *target;  {
	static char buf[1024];
	FILE *fp;
	
	sprintf(buf,"cd %s; uudecode",target);

	fp=popen(buf,"w");
	if (fp==NULL) return;
	while (fgets(buf,1024,stdin)!=NULL) {
		fprintf(fp,"%s",buf);
		}
	pclose(fp);

	}

void do_lmail_real(from,recp)
char from[];
char recp[]; {
	static char buf[1024];
	FILE *fp;
	long now;
	struct tm *tnow;
	char mid[64];
	
	sprintf(buf,FALLBACK,recp);

	time(&now);
	sprintf(mid,"%ld.%d",now,getpid());
	fp=popen(buf,"w");
	if (fp==NULL) return;
	fprintf(fp,"From %s remote from %s\n",from,nodename);
	fprintf(fp,"Received: by %s (UCG Mailer 2.0)\n",FQDN);
	tnow=(struct tm *)localtime(&now);
	fprintf(fp,"	id %s; %s",mid,asctime(tnow));
	while (fgets(buf,1024,stdin)!=NULL) {
		if (!strncmp(buf,"From ",5)) putc('>',fp);
		fprintf(fp,"%s",buf);
		}
	pclose(fp);
	}

void load_nodename() {
	FILE *ln;
	ln=fopen("/etc/systemid","r");
	fgets(nodename,16,ln);
	fclose(ln);
	nodename[strlen(nodename)-1]=0;
	}

int alias(name)
char *name; {
	FILE *fp;
	int a;
	char abuf[256];
	
	fp=fopen(ALIASES,"r");
	if (fp==NULL) {
#ifdef USE_SYSLOG
		syslog(LOG_ERR,"cannot open %s: %s",ALIASES,strerror(errno));
#endif
		return(2);
		}

	while (fgets(abuf,256,fp)!=NULL) {
		strip_trailing_whitespace(abuf);
		for (a=0; a<strlen(abuf); ++a) {
			if (abuf[a]==',') {
				abuf[a]=0;
				if (!strucmp(name,abuf)) {
					strcpy(name,&abuf[a+1]);
					}
				}
			}
		}
	fclose(fp);
	return(0);
}


void main(argc,argv)
int argc;
char *argv[]; {
	static char from[1024],recp[1024],buf[1024];
	int a,b,eflag;
	int is_test = 0;
	long pos;

#ifdef USE_SYSLOG
	openlog("citmail",LOG_PID,LOG_USER);
#endif
	get_config();
	load_nodename();

	if (!strcmp(argv[1],"-t")) {
		is_test = 1;
#ifdef USE_SYSLOG
		syslog(LOG_NOTICE,"test mode - will not deliver\n");
#endif
		}

	if (is_test == 0) strcpy(recp,argv[1]);
	else strcpy(recp,argv[2]);
#ifdef USE_SYSLOG
	syslog(LOG_NOTICE,"recp: %s",recp);
#endif
	for (a=0; a<2; ++a) {
		alias(recp);
		}

	/* did we alias it back to a remote address? */
	if (	(haschar(recp,'%'))
	   ||	(haschar(recp,'@'))
	   ||	(haschar(recp,'!')) ) {
		sprintf(buf,SENDMAIL,recp);
		system(buf);
		exit(0);
		}

	eflag=0;
	fgets(from,1024,stdin);

	/* strip off the trailing newline */
	strip_trailing_whitespace(from);

	/* first line could be from the the From_ field */
	if (strncmp(from,"From ",5)) {
		sprintf(from,"From someone@%s ",config.c_fqdn);
		}

	else {

		/* strip off the word From */
		strcpy(from,&from[5]);
	
		/* now strip off leading blanks */
		while (from[0]==' ') strcpy(from,&from[1]);
	
		/* collapse "remote from" message into the bang path */
		for (a=0; a<strlen(from); ++a)
			if (!strncmp(&from[a],"remote from ",12)) {
				strcpy(buf,&from[a+12]);
				from[a-1]=0;
				strcat(buf,"!"); strcat(buf,from);
				strcpy(from,buf);
				}
		}

	/* strip our system name out, if it's in there multiple times */
	while ( (!struncmp(from,nodename,strlen(nodename)))
		&& (from[strlen(nodename)]=='!') ) 
			strcpy(from,&from[strlen(nodename)+1]);

	b=islocalok(recp,&pos);

#ifdef USE_SYSLOG
	syslog(LOG_NOTICE,"sender is <%s>",from);
#endif

	/* various ways we can deliver mail... */
	if (!strcmp(recp,"uudecode")) {
#ifdef USE_SYSLOG
		syslog(LOG_NOTICE,"uudecoding to loading dock");
#endif
		if (is_test == 0) do_uudecode(LOADINGDOCK);
		}
	if (!strcmp(recp,"cit86net")) {
#ifdef USE_SYSLOG
		syslog(LOG_NOTICE,"uudecoding to cit86net spool");
#endif
		if (is_test == 0) {
			do_uudecode(CIT86NET);
			system("exec ./BatchTranslate86");
			}
		}
	else if (!strcmp(recp,"null")) {
#ifdef USE_SYSLOG
		syslog(LOG_NOTICE,"zapping nulled message");
#endif
		}
	else if (!struncmp(recp,"room_",5)) {
#ifdef USE_SYSLOG
		syslog(LOG_NOTICE,"to room %s",recp);
#endif
		if (is_test == 0) do_roommail(from,recp);
		}
	else if (b==1) {
#ifdef USE_SYSLOG
		syslog(LOG_NOTICE,"to sys user %s",recp);
#endif
		if (is_test == 0) do_lmail_real(from,recp);
		}
	else if (b>1) {
#ifdef USE_SYSLOG
		syslog(LOG_NOTICE,"to citadel user %s",recp);
#endif
		if (is_test == 0) do_citmail(from,recp,pos,LOCAL);
		}
	else {
		/* should be able to bounce if undeliverable */
		if (is_test == 0) bounce(from,argv[1]);
		}
#ifdef USE_SYSLOG
	closelog();
#endif
	execlp("./netproc","netproc",NULL);
	exit(0);
	}

